/*
 * $Id: nvidia.cpp,v 1.31 2004/01/09 18:20:29 giessl Exp $
 *
 * nVIDIA KWin client (version 0.8)
 *
 * Copyright (C) 2002 Fredrik H�lund <fredrik@kde.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <kconfig.h>
#include <klocale.h>
#include <kiconeffect.h>

#include <qpainter.h>
#include <qlayout.h>
#include <qbitmap.h>
#include <qstyle.h>
#include <qtooltip.h>
#include <qwidget.h>
#include <qlabel.h>

#include <X11/Xlib.h>

#include "nvidia.h"
#include "nvidia.moc"



// -------------------------------------------------------------------------------------------



namespace nVIDIA
  {

  const int buttonMargin     =  11;  // Margin between the window edge and the buttons
  const int buttonSpacing    =  3;  // Spacing between the titlebar buttons
  const int iconSpacing      =  4;  // Spacing between the icon and the text label

  // Default button layout
  const char default_left[]  = "M";
  const char default_right[] = "HIAX";

  nVIDIAHandler *clientHandler = NULL;
  bool nvidia_initialized = false;



  // -------------------------------------------------------------------------------------------



  nVIDIAHandler::nVIDIAHandler()
  {
    for ( int i = 0; i < NumTiles; i++ )
      {
        activeTiles[i]   = NULL;
        inactiveTiles[i] = NULL;
      }
    for ( int b = 0; b < NumButton; b++ )
      {
        activeButtons[b]   = NULL;
        inactiveButtons[b] = NULL;
      }   
    for ( int e = 0; e < NumButtonDecos; e++ )
      {
        overbuttonDecos[e]   = NULL;
        pressedbuttonDecos[e] = NULL;
        inoverbuttonDecos[e]   = NULL;
        inpressedbuttonDecos[e] = NULL;      
        activebuttonDecos[e]   = NULL;
        inactivebuttonDecos[e] = NULL;
      }

    settings_cache = NULL;

    imageDb = nVIDIAImageDb::instance();

    readConfig();
    createPixmaps();

    nvidia_initialized = true;
  }


  nVIDIAHandler::~nVIDIAHandler()
  {
    nvidia_initialized = false;
    destroyPixmaps();

    delete settings_cache;

    nVIDIAImageDb::release();
    imageDb = NULL;
    clientHandler = NULL;
  }


  void nVIDIAHandler::createPixmaps()
  {
    int heightOffset;
    int widthOffset;
    switch(options()->preferredBorderSize(this))
      {
      case BorderLarge:
        widthOffset = 1;
        heightOffset = 0;
        break;
      case BorderVeryLarge:
        widthOffset = 1;
        heightOffset = 0;
        break;
      case BorderHuge:
        widthOffset = 1;
        heightOffset = 0;
        break;
      case BorderVeryHuge:
        widthOffset = 1;
        heightOffset = 10;
        break;
      case BorderOversized:
        widthOffset = 1;
        heightOffset = 25;
        break;
      case BorderTiny:
      case BorderNormal:
      default:
        widthOffset = 0;
        heightOffset = 0;
      }
    int fontHeight = QFontMetrics(options()->font(true)).height();
    if (fontHeight > heightOffset + 20)
      heightOffset = fontHeight - 20;

    QString size = (heightOffset < 8) ? "" : (heightOffset < 20) ? "-large" : "-huge";

    QColor titleColor, captionColor, buttonColor;
    QImage *titleCenter = NULL, *captionLeft = NULL,
                          *captionRight = NULL, *captionCenter = NULL;


    // Active tiles
    // -------------------------------------------------------------------------
    captionColor = KDecoration::options()->color( ColorTitleBar,   true );
    titleColor   = KDecoration::options()->color( ColorTitleBlend, true );

    // Load the titlebar corners.
    activeTiles[ TitleLeft ]  = loadPixmap( "titlebar-left-act");//,  titleColor );
    activeTiles[ TitleRight ] = loadPixmap( "titlebar-right-act");//, titleColor );

    // Load the titlebar center tile image (this will be used as
    //     the background for the caption bubble tiles).
    titleCenter = loadImage( "titlebar-center-act");//, titleColor );

    // Load the small version of the caption bubble corner & center images.
    captionLeft   = loadImage( "caption-small-left-act");//,   captionColor );
    captionRight  = loadImage( "caption-small-right-act");//,  captionColor );
    captionCenter = loadImage( "caption-small-center-act");//, captionColor );

    // Create the caption bubble tiles (by blending the images onto the titlebar)
    activeTiles[ CaptionSmallLeft   ] = composite( captionLeft,   titleCenter );
    activeTiles[ CaptionSmallRight  ] = composite( captionRight,  titleCenter );
    activeTiles[ CaptionSmallCenter ] = composite( captionCenter, titleCenter );

    delete captionLeft;
    delete captionRight;
    delete captionCenter;

    // Now do the same with the large version
    captionLeft   = loadImage( "caption-large-left");//,   captionColor );
    captionRight  = loadImage( "caption-large-right");//,  captionColor );
    captionCenter = loadImage( "caption-large-center");//, captionColor );

    activeTiles[ CaptionLargeLeft   ] = composite( captionLeft,   titleCenter );
    activeTiles[ CaptionLargeRight  ] = composite( captionRight,  titleCenter );
    activeTiles[ CaptionLargeCenter ] = composite( captionCenter, titleCenter );

    delete captionLeft;
    delete captionRight;
    delete captionCenter;

    // Create the titlebar center tile
    activeTiles[ TitleCenter ] = new QPixmap( *titleCenter );

    delete titleCenter;

    // Load the left & right border pixmaps
    activeTiles[ BorderLeft ]  = loadPixmap( "border-left");//,  titleColor );
    activeTiles[ BorderRight ] = loadPixmap( "border-right");//, titleColor );

    // Load the bottom grabbar pixmaps
    if ( largeGrabBars )
      {
        activeTiles[ GrabBarLeft ]   = loadPixmap( "grabbar-left");//,   titleColor );
        activeTiles[ GrabBarRight ]  = loadPixmap( "grabbar-right");//,  titleColor );
        activeTiles[ GrabBarCenter ] = loadPixmap( "grabbar-center");//, titleColor );
      }
    else
      {
        activeTiles[ GrabBarLeft ]   = loadPixmap( "bottom-left");//,   titleColor );
        activeTiles[ GrabBarRight ]  = loadPixmap( "bottom-right");//,  titleColor );
        activeTiles[ GrabBarCenter ] = loadPixmap( "bottom-center");//, titleColor );
      }

    // Inactive tiles
    // -------------------------------------------------------------------------
    captionColor = KDecoration::options()->color( ColorTitleBar,   false );
    titleColor   = KDecoration::options()->color( ColorTitleBlend, false );

    inactiveTiles[ TitleLeft ]  = loadPixmap( "titlebar-left-dis");//,  titleColor );
    inactiveTiles[ TitleRight ] = loadPixmap( "titlebar-right-dis");//, titleColor );

    titleCenter = loadImage( "titlebar-center-dis");//, titleColor );

    captionLeft   = loadImage( "caption-small-left-dis");//,   captionColor );
    captionRight  = loadImage( "caption-small-right-dis");//,  captionColor );
    captionCenter = loadImage( "caption-small-center-dis");//, captionColor );

    inactiveTiles[ CaptionSmallLeft  ]  = composite( captionLeft,   titleCenter );
    inactiveTiles[ CaptionSmallRight ]  = composite( captionRight,  titleCenter );
    inactiveTiles[ CaptionSmallCenter ] = composite( captionCenter, titleCenter );

    delete captionLeft;
    delete captionRight;
    delete captionCenter;

    inactiveTiles[ TitleCenter ] = new QPixmap( *titleCenter );

    delete titleCenter;

    inactiveTiles[ BorderLeft ]  = loadPixmap( "border-left");//,  titleColor );
    inactiveTiles[ BorderRight ] = loadPixmap( "border-right");//, titleColor );

    if ( largeGrabBars )
      {
        inactiveTiles[ GrabBarLeft ]   = loadPixmap( "grabbar-left");//,   titleColor );
        inactiveTiles[ GrabBarRight ]  = loadPixmap( "grabbar-right");//,  titleColor );
        inactiveTiles[ GrabBarCenter ] = loadPixmap( "grabbar-center");//, titleColor );
      }
    else
      {
        inactiveTiles[ GrabBarLeft ]   = loadPixmap( "bottom-left");//,   titleColor );
        inactiveTiles[ GrabBarRight ]  = loadPixmap( "bottom-right");//,  titleColor );
        inactiveTiles[ GrabBarCenter ] = loadPixmap( "bottom-center");//, titleColor );
      }

    //active Buttons
    // -------------------------------------------------------------------------
    // buttonColor  =  KDecoration::options()->color( ButtonBg, true );
    activeButtons[ ButtonRound ] = loadPixmap("titlebutton-round-act"+size);
    activeButtons[ ButtonSquare] = loadPixmap("titlebutton-square-act"+size);
    //inactive Buttons
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    inactiveButtons[ ButtonRound ] = loadPixmap("titlebutton-round-dis"+size);
    inactiveButtons[ ButtonSquare ] = loadPixmap("titlebutton-square-dis"+size);
    //active Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    activebuttonDecos[Menu] = loadPixmap("menu-act");
    activebuttonDecos[OnAllDesktops] = loadPixmap("alldesktop-act");
    activebuttonDecos[NotOnAllDesktops] = loadPixmap("noalldesktop-act");
    activebuttonDecos[Help] = loadPixmap("help-act");
    activebuttonDecos[Minimize] = loadPixmap("minimize-act");
    activebuttonDecos[Maximize] = loadPixmap("maximize-act");
    activebuttonDecos[Restore] = loadPixmap("restore-act");
    activebuttonDecos[Close] = loadPixmap("close-act");
    //inactive Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    inactivebuttonDecos[Menu] = loadPixmap("menu-dis");
    inactivebuttonDecos[OnAllDesktops] = loadPixmap("alldesktop-dis");
    inactivebuttonDecos[NotOnAllDesktops] = loadPixmap("noalldesktop-dis");
    inactivebuttonDecos[Help] = loadPixmap("help-dis");
    inactivebuttonDecos[Minimize] = loadPixmap("minimize-dis");
    inactivebuttonDecos[Maximize] = loadPixmap("maximize-dis");
    inactivebuttonDecos[Restore] = loadPixmap("restore-dis");
    inactivebuttonDecos[Close] = loadPixmap("close-dis");
    //over active Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    overbuttonDecos[Minimize] = loadPixmap("minimize-act-over");
    overbuttonDecos[Maximize] = loadPixmap("maximize-act-over");
    overbuttonDecos[Restore] = loadPixmap("restore-act-over");
    overbuttonDecos[Close] = loadPixmap("close-act-over");
    //pressed active Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    pressedbuttonDecos[Minimize]  = loadPixmap("minimize-act-pressed");
    pressedbuttonDecos[Maximize] = loadPixmap("maximize-act-pressed");
    pressedbuttonDecos[Restore] = loadPixmap("restore-act-pressed");
    pressedbuttonDecos[Close] = loadPixmap("close-act-pressed");
    //over active Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    inoverbuttonDecos[Minimize] = loadPixmap("minimize-dis-over");
    inoverbuttonDecos[Maximize] = loadPixmap("maximize-dis-over");
    inoverbuttonDecos[Restore] = loadPixmap("restore-dis-over");
    inoverbuttonDecos[Close] = loadPixmap("close-dis-over");
    //pressed active Buttons deco
    // -------------------------------------------------------------------------
    //  buttonColor  =  KDecoration::options()->color( ButtonBg, false );
    inpressedbuttonDecos[Minimize]  = loadPixmap("minimize-dis-pressed");
    inpressedbuttonDecos[Maximize] = loadPixmap("maximize-dis-pressed");
    inpressedbuttonDecos[Restore] = loadPixmap("restore-dis-pressed");
    inpressedbuttonDecos[Close] = loadPixmap("close-dis-pressed");
    // Prepare the tiles for use
    // -------------------------------------------------------------------------
    if ( QApplication::reverseLayout() )
      {

        // Fix lighting
        flip( activeTiles[CaptionSmallLeft], activeTiles[CaptionSmallRight] );
        flip( inactiveTiles[CaptionSmallLeft], inactiveTiles[CaptionSmallRight] );

        flip( activeTiles[CaptionLargeLeft], activeTiles[CaptionLargeRight] );

        flip( activeTiles[TitleLeft], activeTiles[TitleRight] );
        flip( inactiveTiles[TitleLeft], inactiveTiles[TitleRight] );

        flip( activeTiles[BorderLeft], activeTiles[BorderRight] );
        flip( inactiveTiles[BorderLeft], inactiveTiles[BorderRight] );

        flip( activeTiles[GrabBarLeft], activeTiles[GrabBarRight] );
        flip( inactiveTiles[GrabBarLeft], inactiveTiles[GrabBarRight] );

        flip( activeButtons[ ButtonRound ], activeButtons[ ButtonSquare ] );
        flip( inactiveButtons[ ButtonRound ], inactiveButtons[ ButtonSquare ] );
      }

    // Pretile the center & border tiles for optimal performance
    pretile( activeTiles[ CaptionSmallCenter ], 64, Qt::Horizontal );
    pretile( activeTiles[ CaptionLargeCenter ], 64, Qt::Horizontal );
    pretile( activeTiles[ TitleCenter ], 64, Qt::Horizontal );
    pretile( activeTiles[ GrabBarCenter ], 128, Qt::Horizontal );
    pretile( activeTiles[ BorderLeft ], 128, Qt::Vertical );
    pretile( activeTiles[ BorderRight ], 128, Qt::Vertical );

    pretile( inactiveTiles[ CaptionSmallCenter ], 64, Qt::Horizontal );
    pretile( inactiveTiles[ TitleCenter ], 64, Qt::Horizontal );
    pretile( inactiveTiles[ GrabBarCenter ], 128, Qt::Horizontal );
    pretile( inactiveTiles[ BorderLeft ], 128, Qt::Vertical );
    pretile( inactiveTiles[ BorderRight ], 128, Qt::Vertical );

    if (heightOffset > 0)
      {
        addHeight (heightOffset, activeTiles[TitleLeft]);
        addHeight (heightOffset, activeTiles[TitleCenter]);
        addHeight (heightOffset, activeTiles[TitleRight]);
        addHeight (heightOffset, activeTiles[CaptionSmallLeft]);
        addHeight (heightOffset, activeTiles[CaptionSmallCenter]);
        addHeight (heightOffset, activeTiles[CaptionSmallRight]);
        addHeight (heightOffset, activeTiles[CaptionLargeLeft]);
        addHeight (heightOffset, activeTiles[CaptionLargeCenter]);
        addHeight (heightOffset, activeTiles[CaptionLargeRight]);

        addHeight (heightOffset, inactiveTiles[TitleLeft]);
        addHeight (heightOffset, inactiveTiles[TitleCenter]);
        addHeight (heightOffset, inactiveTiles[TitleRight]);
        addHeight (heightOffset, inactiveTiles[CaptionSmallLeft]);
        addHeight (heightOffset, inactiveTiles[CaptionSmallCenter]);
        addHeight (heightOffset, inactiveTiles[CaptionSmallRight]);
      }

    if (widthOffset > 0)
      {
        addWidth (widthOffset, activeTiles[BorderLeft], true, activeTiles[GrabBarCenter]);
        addWidth (widthOffset, activeTiles[BorderRight], false, activeTiles[GrabBarCenter]);
        addWidth (widthOffset, inactiveTiles[BorderLeft], true, inactiveTiles[GrabBarCenter]);
        addWidth (widthOffset, inactiveTiles[BorderRight], false, inactiveTiles[GrabBarCenter]);

        if (largeGrabBars)
          widthOffset = widthOffset*3/2;

        addHeight (widthOffset, activeTiles[GrabBarLeft]);
        addHeight (widthOffset, activeTiles[GrabBarCenter]);
        addHeight (widthOffset, activeTiles[GrabBarRight]);
        addHeight (widthOffset, inactiveTiles[GrabBarLeft]);
        addHeight (widthOffset, inactiveTiles[GrabBarCenter]);
        addHeight (widthOffset, inactiveTiles[GrabBarRight]);
      }
  }



  void nVIDIAHandler::destroyPixmaps()
  {
    for ( int i = 0; i < NumTiles; i++ )
      {
        delete activeTiles[i];
        delete inactiveTiles[i];
        activeTiles[i] = NULL;
        inactiveTiles[i] = NULL;
      }
    for ( int a = 0; a < NumButton; a++ )
      {
        delete activeButtons[a];
        delete inactiveButtons[a];
        activeButtons[a] = NULL;
        inactiveButtons[a] = NULL;
      }
    for ( int c = 0; c < NumButtonDecos; c++)
      {
        delete activebuttonDecos[c];
        delete inactivebuttonDecos[c];
        activebuttonDecos[c] = NULL;
        inactivebuttonDecos[c] = NULL;   
        delete overbuttonDecos[c];
        delete pressedbuttonDecos[c];
        delete inoverbuttonDecos[c];
        delete inpressedbuttonDecos[c];
        overbuttonDecos[c] = NULL;
        pressedbuttonDecos[c] = NULL;
        inoverbuttonDecos[c] = NULL;
        inpressedbuttonDecos[c] = NULL;
      }
  }


  void nVIDIAHandler::addWidth (int width, QPixmap *&pix, bool left, QPixmap *bottomPix)
  {
    int w = pix->width()+width;
    int h = pix->height();

    QPixmap *tmp = new QPixmap (w, h);
    tmp->fill ();
    QPainter p;
    p.begin (tmp);

    for (int i = 0; i < h; i++)
      p.drawPixmap (0, i, *bottomPix, i%2, 0, w,1);

    if (left)
      p.drawPixmap(0, 0, *pix);
    else
      p.drawPixmap(width, 0, *pix);

    p.end();

    delete pix;
    pix = tmp;
  }


  void nVIDIAHandler::addHeight (int height, QPixmap *&pix)
  {
    int w = pix->width();
    int h = pix->height()+height;

    QPixmap *tmp = new QPixmap (w, h);
    QPainter p;
    p.begin (tmp);
    if (pix->height() > 10)
      {
        p.drawPixmap(0, 0, *pix, 0, 0, w, 11);
        for (int i = 0; i < height; i+=2)
          p.drawPixmap(0, 11+i, *pix, 0, 11, w, 2);
        p.drawPixmap(0, 11+height, *pix, 0, 11, w, -1);
      }
    else
      {
        int lines  = h-3;
        int factor = pix->height()-3;
        for (int i = 0; i < lines; i++)
          p.drawPixmap(0, i, *pix, 0, i*factor/lines, w, 1);
        p.drawPixmap(0, lines, *pix, 0, factor, w, 3);
      }
    p.end();

    delete pix;
    pix = tmp;
  }


  void nVIDIAHandler::flip( QPixmap *&pix1, QPixmap *&pix2 )
  {
    // Flip the pixmaps horizontally
    QPixmap *tmp = new QPixmap( pix1->xForm( QWMatrix(-1,0,0,1,pix1->width(),0) ) );

    delete pix1;
    pix1 = new QPixmap( pix2->xForm( QWMatrix(-1,0,0,1,pix2->width(),0) ) );

    delete pix2;
    pix2 = tmp;
  }


  void nVIDIAHandler::flip( QPixmap *&pix )
  {
    // Flip the pixmap horizontally
    QPixmap *tmp = new QPixmap( pix->xForm( QWMatrix(-1,0,0,1,pix->width(),0) ) );
    delete pix;
    pix = tmp;
  }


  void nVIDIAHandler::pretile( QPixmap *&pix, int size, Qt::Orientation dir )
  {
    QPixmap *newpix;
    QPainter p;

    if ( dir == Qt::Horizontal )
      newpix = new QPixmap( size, pix->height() );
    else
      newpix = new QPixmap( pix->width(), size );

    p.begin( newpix );
    p.drawTiledPixmap( newpix->rect(), *pix ) ;
    p.end();

    delete pix;
    pix = newpix;
  }


  void nVIDIAHandler::readConfig()
  {
    KConfig *c = new KConfig( "kwinnvidiarc" );

    c->setGroup( "General" );
    showIcons = c->readBoolEntry( "ShowAppIcons", true );
    shadowedText = c->readBoolEntry( "UseShadowedText", true );
    smallCaptionBubbles = c->readBoolEntry( "SmallCaptionBubbles", true );
    largeGrabBars = c->readBoolEntry( "LargeGrabBars", false );

    if ( ! settings_cache )
      {
        settings_cache = new SettingsCache;
        settings_cache->largeGrabBars = largeGrabBars;
        settings_cache->smallCaptionBubbles = smallCaptionBubbles;
      }

    delete c;
  }


  QPixmap *nVIDIAHandler::composite( QImage *over, QImage *under )
  {
    QImage dest( over->width(), over->height(), 32 );
    int width = over->width(), height = over->height();

    // Clear the destination image
    Q_UINT32 *data = reinterpret_cast<Q_UINT32*>( dest.bits() );
    for (int i = 0; i < width * height; i++)
      *(data++) = 0;

    // Copy the under image (bottom aligned) to the destination image
    for (int y1 = height - under->height(), y2 = 0; y1 < height; y1++, y2++ )
      {
        register Q_UINT32 *dst = reinterpret_cast<Q_UINT32*>( dest.scanLine(y1) );
        register Q_UINT32 *src = reinterpret_cast<Q_UINT32*>( under->scanLine(y2) );

        for ( int x = 0; x < width; x++ )
          *(dst++) = *(src++);
      }

    // Blend the over image onto the destination
    register Q_UINT32 *dst = reinterpret_cast<Q_UINT32*>( dest.bits() );
    register Q_UINT32 *src = reinterpret_cast<Q_UINT32*>( over->bits() );
    for ( int i = 0; i < width * height; i++ )
      {
        int r1 = qRed( *dst ), g1 = qGreen( *dst ), b1 = qBlue( *dst );
        int r2 = qRed( *src ), g2 = qGreen( *src ), b2 = qBlue( *src );
        int a  = qAlpha( *src );

        if ( a == 0xff )
          *dst = *src;

        else if ( a != 0x00 )
          *dst = qRgba( Q_UINT8( r1 + (((r2 - r1) * a) >> 8) ),
                        Q_UINT8( g1 + (((g2 - g1) * a) >> 8) ),
                        Q_UINT8( b1 + (((b2 - b1) * a) >> 8) ),
                        0xff );

        else if ( qAlpha(*dst) == 0x00 )
          *dst = 0;

        src++;
        dst++;
      }

    // Create the final pixmap and return it
    return new QPixmap( dest );
  }


  QImage *nVIDIAHandler::loadImage( const QString &name)//, const QColor &col )
  {
    // if ( col.isValid() )
    //{
    QImage *img = new QImage( imageDb->image(name)->copy() );
    //  KIconEffect::colorize( *img, col, 1.0 );
    return img;
    // }
    //else
    // return new QImage( imageDb->image(name)->copy() );
  }


  QPixmap *nVIDIAHandler::loadPixmap( const QString &name)//, const QColor &col )
  {
    QImage *img = loadImage( name);//, col );
    QPixmap *pix = new QPixmap( *img );
    delete img;

    return pix;
  }


  bool nVIDIAHandler::reset( unsigned long changed )
  {
    nvidia_initialized = false;

    bool needHardReset  = false;
    bool pixmapsInvalid = false;

    // Re-read the config file
    readConfig();

    if ( changed & SettingBorder )
      {
        pixmapsInvalid = true;
        needHardReset = true;
      }
    if ( changed & SettingFont )
      {
        pixmapsInvalid = true;
        needHardReset = true;
      }
    // Check if the color scheme has changed
    if ( changed & SettingColors )
      {
        pixmapsInvalid = true;
      }
    // Check if button positions have changed

    if ( changed & SettingButtons )
      {
        needHardReset = true;
      }

    // Check if tooltips options have changed
    if ( changed & SettingTooltips )
      {
        needHardReset = true;
      }

    if ( (settings_cache->largeGrabBars != largeGrabBars) )
      {
        pixmapsInvalid = true;
        needHardReset = true;
      }

    if ( (settings_cache->smallCaptionBubbles != smallCaptionBubbles) )
      {
        needHardReset = true;
      }

    // Update our config cache
    settings_cache->largeGrabBars       = largeGrabBars;
    settings_cache->smallCaptionBubbles = smallCaptionBubbles;

    // Do we need to recreate the pixmaps?
    if ( pixmapsInvalid )
      {
        destroyPixmaps();
        createPixmaps();
      }

    nvidia_initialized = true;

    // Do we need to "hit the wooden hammer" ?
    if ( !needHardReset )
      resetDecorations( changed );
    return needHardReset;
  }


  const QPixmap *nVIDIAHandler::tile( TilePixmap tilePix, bool active ) const
    {
      return ( active ? activeTiles[ tilePix ] : inactiveTiles[ tilePix ] );
    }

  const QPixmap *nVIDIAHandler::btile(TileButtons buttonPix, bool active) const
    {
      return (active ? activeButtons[ buttonPix ] : inactiveButtons[ buttonPix ]);
    }

  const QPixmap *nVIDIAHandler::decotile(ButtonDeco tiledeco, bool active) const
    {
      return (active ? activebuttonDecos[ tiledeco ] : inactivebuttonDecos[ tiledeco ]);
    }

  const QPixmap *nVIDIAHandler::decostate(ButtonDeco decostate, bool active, bool state) const
    {
      if(active)
        {
          return (state ? pressedbuttonDecos[ decostate ] : overbuttonDecos[ decostate ]);//true means pressed:false means over
        }
      else
        {
          return (state ? inpressedbuttonDecos[ decostate ] : inoverbuttonDecos[ decostate ]);
        }
    }

  KDecoration* nVIDIAHandler::createDecoration( KDecorationBridge* bridge )
  {
    return new nVIDIAClient( bridge, this );
  }

  QValueList< nVIDIAHandler::BorderSize > nVIDIAHandler::borderSizes() const
    { // the list must be sorted
      return QValueList< BorderSize >() << BorderNormal << BorderLarge <<
             BorderVeryLarge <<  BorderHuge << BorderVeryHuge << BorderOversized;
    }


  // -------------------------------------------------------------------------------------------



  nVIDIAButton::nVIDIAButton( nVIDIAClient* c, const char *name, Button btn, const QString &tip, const int realizeBtns )
      : QButton( c->widget(), name ),
      client( c ), button( btn ), hover( false ), lastbutton( 0 )
  {
    realizeButtons = realizeBtns;
    bool active = client->isActive();
    QToolTip::add
      ( this, tip ); // FRAME
    setBackgroundMode( NoBackground );
    setCursor( arrowCursor );
    if (name == "menu" || name == "on_all_desktops" || name == "help" )
      {
        setAutoMask(true);
        int size = clientHandler->btile( ButtonRound, active )->height();
        setFixedSize( size, size );
      }
    else
      {
        int size = clientHandler->btile( ButtonSquare, active )->height();
        setFixedSize( size, size );
      }
    setToggleButton( (button == OnAllDesktopsButton) );
  }


  nVIDIAButton::~nVIDIAButton()
  {
    // Empty.
  }


  void nVIDIAButton::enterEvent( QEvent *e )
  {
    QButton::enterEvent( e );

    hover = true;
    repaint( false );
  }


  void nVIDIAButton::leaveEvent( QEvent *e )
  {
    QButton::leaveEvent( e );

    hover = false;
    repaint( false );
  }


  void nVIDIAButton::mousePressEvent( QMouseEvent *e )
  {
    lastbutton = e->button();
    QMouseEvent me( e->type(), e->pos(), e->globalPos(), (e->button()&realizeButtons)?LeftButton:NoButton, e->state() );
    QButton::mousePressEvent( &me );
  }


  void nVIDIAButton::mouseReleaseEvent( QMouseEvent *e )
  {
    lastbutton = e->button();
    QMouseEvent me( e->type(), e->pos(), e->globalPos(), (e->button()&realizeButtons)?LeftButton:NoButton, e->state() );
    QButton::mouseReleaseEvent( &me );
  }


  void nVIDIAButton::drawButton( QPainter *p )
  {
    const QPixmap *pix;
    int size;
    bool active =client->isActive();
    QString state;
    state = "normal";
    const QPixmap *deco;
    // Get the bevel from the client handler
    if ( button == MenuButton || button == OnAllDesktopsButton || button == HelpButton )
      {
        pix = clientHandler->btile( ButtonRound, active );
        // Draw the button round background
        size = clientHandler->btile( ButtonRound, active )->height();
      }
    else
      {
        pix = clientHandler->btile( ButtonSquare, active );
        // Draw the button square background
        size = clientHandler->btile( ButtonSquare, active )->height();
      }

    if ( isDown() )
      {
        // Pressed
        if ( button == MenuButton || button == OnAllDesktopsButton || button == HelpButton )
          {
            state = "pressed";
            p->drawPixmap( QPoint(), *pix, QStyle::visualRect( QRect(2*size, 0, size, size), pix->rect() ) );
            p->translate( QApplication::reverseLayout() ? -1 : 1,  1 );
          }
        else
          {
            state = "pressed";
            p->drawPixmap( QPoint(), *pix, QStyle::visualRect( QRect(2*size+iconSpacing, 0, size, size), pix->rect() ) );
            p->translate( QApplication::reverseLayout() ? -1 : 1,  1 );
          }
      }
    else if ( hover )
      // Mouse over
      if ( button == MenuButton || button == OnAllDesktopsButton || button == HelpButton )
        {
          state= "over";
          p->drawPixmap( QPoint(), *pix, QStyle::visualRect( QRect(size, 0, size, size), pix->rect() ) );
        }
      else
        {
          state= "over";
          p->drawPixmap( QPoint(), *pix, QStyle::visualRect( QRect(size+iconSpacing-2, 0, size, size), pix->rect() ) );
        }
    else
      // Normal
      p->drawPixmap( QPoint(), *pix, QStyle::visualRect( QRect(0, 0, size, size), pix->rect() ) );


    // Draw the button deco on the bevel
    switch ( button )
      {
      case MenuButton:
        deco = clientHandler->decotile(Menu, active);
        break;

      case OnAllDesktopsButton:
        deco = clientHandler->decotile( isOn() ? NotOnAllDesktops : OnAllDesktops, active );
        break;

      case HelpButton:
        deco = clientHandler->decotile( Help, active );
        if ( QApplication::reverseLayout() )
          p->translate( 2, 0 );
        break;

      case MinButton:
        if (state == "pressed")
          {
            deco = clientHandler->decostate(Minimize, active, true);
          }
        else if (state == "over")
          {
            deco = clientHandler->decostate(Minimize, active, false);
          }
        else
          deco = clientHandler->decotile( Minimize, active );
        break;

      case MaxButton:
        if (state == "pressed")
          {
            deco = clientHandler->decostate(client->maximizeMode() == nVIDIAClient::MaximizeFull ? Restore : Maximize, active, true);
          }
        else if (state == "over")
          {
            deco = clientHandler->decostate(client->maximizeMode() == nVIDIAClient::MaximizeFull ? Restore : Maximize, active, false);
          }
        else
          deco = clientHandler->decotile( client->maximizeMode() == nVIDIAClient::MaximizeFull ? Restore : Maximize, active );
        break;

      case CloseButton:
        if (state == "pressed")
          {
            deco = clientHandler->decostate(Close, active, true);
          }
        else if (state == "over")
          {
           deco = clientHandler->decostate(Close, active, false);
          }
        else
          deco = clientHandler->decotile( Close, active );
        break;

      default:
        deco = NULL;
      }
    p->drawPixmap( (size-clientHandler->decotile( Close, active )->width()+1)/2, (size-clientHandler->decotile( Close, active )->height())/2, *deco );
  }



  // ------------------------------------------------------------------------------------------



  nVIDIAClient::nVIDIAClient( KDecorationBridge* bridge, KDecorationFactory* factory )
      : KDecoration( bridge, factory ),
      activeIcon( NULL ), inactiveIcon( NULL ), captionBufferDirty( true ), maskDirty( true )
{}

  void nVIDIAClient::init()
  {
    createMainWidget( WStaticContents | WResizeNoErase | WRepaintNoErase );
    widget()->installEventFilter( this );

    // Minimize flicker
    widget()->setBackgroundMode( NoBackground );

    for ( int i=0; i < NumButtons; i++ )
      button[i] = NULL;

    createLayout();
  }

  void nVIDIAClient::createLayout()
  {

    QVBoxLayout *mainLayout   = new QVBoxLayout( widget() );
    QBoxLayout *titleLayout   = new QBoxLayout( 0, QBoxLayout::LeftToRight, 0, 0, 0 );
    QHBoxLayout *windowLayout = new QHBoxLayout();

    largeTitlebar = ( !maximizedVertical() && clientHandler->largeCaptionBubbles() );
    largeCaption = ( isActive() && largeTitlebar );

    int grabBarHeight = clientHandler->grabBarHeight();
    int topSpacing = ( largeTitlebar ? 4 : 1 );
    int leftBorderWidth    = clientHandler->tile( BorderLeft, true )->width();
    int rightBorderWidth   = clientHandler->tile( BorderRight, true )->width();
    topSpacer = new QSpacerItem( 10, topSpacing,
                                 QSizePolicy::Expanding, QSizePolicy::Minimum );

    mainLayout->addItem( topSpacer );

    mainLayout->addLayout( titleLayout );     // Titlebar
    mainLayout->addLayout( windowLayout, 1 ); // Left border + window + right border
    mainLayout->addSpacing( grabBarHeight );  // Bottom grab bar

    titleLayout->setSpacing( buttonSpacing );

    titleLayout->addSpacing( buttonMargin );      // Left button margin
    addButtons( titleLayout, options()->customButtonPositions() ?
                options()->titleButtonsLeft() : QString(default_left) );

    titlebar = new QSpacerItem( 10, clientHandler->titleBarHeight(largeTitlebar)
                                - topSpacing, QSizePolicy::Expanding, QSizePolicy::Minimum );
    titleLayout->addItem( titlebar );

    titleLayout->addSpacing( buttonSpacing );
    addButtons( titleLayout, options()->customButtonPositions() ?
                options()->titleButtonsRight() : QString(default_right) );
    titleLayout->addSpacing( buttonMargin - 1 );  // Right button margin

    windowLayout->addSpacing( leftBorderWidth );                // Left border
    if( isPreview())
      windowLayout->addWidget( new QLabel( i18n( "<center><b>nVIDIA</b></center>" ), widget()));
    else
      windowLayout->addItem( new QSpacerItem( 0, 0 )); //no widget in the middle
    windowLayout->addSpacing( rightBorderWidth );                // Right border

  }


  nVIDIAClient::~nVIDIAClient()
  {
    delete activeIcon;
    delete inactiveIcon;

    activeIcon = inactiveIcon = NULL;
  }


  void nVIDIAClient::reset( unsigned long )
  {
    if ( clientHandler->largeCaptionBubbles() && !largeTitlebar )
      {
        // We're switching from small caption bubbles to large
        if ( !maximizedVertical() )
          {
            topSpacer->changeSize( 10, 4, QSizePolicy::Expanding, QSizePolicy::Minimum );
            largeTitlebar = true;
            largeCaption = isActive();

            widget()->layout()->activate();

            // Compensate for the titlebar size change

            // TODO This is wrong, this may break size increments (see bug #53784).
            // FRAME
            widget()->setGeometry( widget()->x(), widget()->y() - 3, width(), height() + 3 );
          }
      }
    else if ( !clientHandler->largeCaptionBubbles() && largeTitlebar )
      {
        // We're switching from large caption bubbles to small
        topSpacer->changeSize( 10, 1, QSizePolicy::Expanding, QSizePolicy::Minimum );
        largeTitlebar = largeCaption = false;

        widget()->layout()->activate();

        // Compensate for the titlebar size change
        // FRAME
        widget()->setGeometry( widget()->x(), widget()->y() + 3, width(), height() - 3 );
      }

    calculateCaptionRect();

    captionBufferDirty = maskDirty = true;

    // Only repaint the window if it's visible
    // (i.e. not minimized and on the current desktop)
    if ( widget()->isVisible() )
      {
        widget()->repaint( false );

        for ( int i = 0; i < NumButtons; i++ )
          if ( button[i] )
            button[i]->repaint( false );
      }

  }


  void nVIDIAClient::addButtons( QBoxLayout *layout, const QString &s )
  {
    for ( uint i=0; i < s.length(); i++ )
      {
        switch ( s[i].latin1() )
          {
            // Menu button
          case 'M' :
            if ( !button[MenuButton] )
              {
                button[MenuButton] = new nVIDIAButton( this, "menu", MenuButton, i18n("Menu"), LeftButton|RightButton );
                connect( button[MenuButton], SIGNAL( pressed() ), SLOT( menuButtonPressed() ) );
                layout->addWidget( button[MenuButton] );
              }
            break;

            // OnAllDesktops button
          case 'S' :
            if ( !button[OnAllDesktopsButton] )
              {
                button[OnAllDesktopsButton] = new nVIDIAButton( this, "on_all_desktops",
                                              OnAllDesktopsButton, isOnAllDesktops()?i18n("Not On All Desktops"):i18n("On All Desktops") );
                if(isOnAllDesktops())
                  button[OnAllDesktopsButton]->toggle();
                connect( button[OnAllDesktopsButton], SIGNAL( clicked() ), SLOT( toggleOnAllDesktops() ) );
                layout->addWidget( button[OnAllDesktopsButton] );
              }
            break;

            // Help button
          case 'H' :
            if ( !button[HelpButton] && providesContextHelp() )
              {
                button[HelpButton] = new nVIDIAButton( this, "help", HelpButton, i18n("Help") );
                connect( button[HelpButton], SIGNAL( clicked() ), SLOT( showContextHelp() ) );
                layout->addWidget( button[HelpButton] );
              }
            break;

            // Minimize button
          case 'I' :
            if ( !button[MinButton] && isMinimizable() )
              {
                button[MinButton] = new nVIDIAButton( this, "minimize", MinButton, i18n("Minimize") );
                connect( button[MinButton], SIGNAL( clicked() ), SLOT( minimize() ) );
                layout->addWidget( button[MinButton] );
              }
            break;

            // Maximize button
          case 'A' :
            if ( !button[MaxButton] && isMaximizable() )
              {
                button[MaxButton] = new nVIDIAButton( this, "maximize", MaxButton, i18n("Maximize"), LeftButton|MidButton|RightButton );
                connect( button[MaxButton], SIGNAL( clicked() ), SLOT( slotMaximize() ) );
                layout->addWidget( button[MaxButton] );
              }
            break;

            // Close button
          case 'X' :
            if ( !button[CloseButton] && isCloseable() )
              {
                button[CloseButton] = new nVIDIAButton( this, "close", CloseButton, i18n("Close") );
                connect( button[CloseButton], SIGNAL( clicked() ), SLOT( closeWindow() ) );
                layout->addWidget( button[CloseButton] );
              }
            break;

            // Additional spacing
          case '_' :
            layout->addSpacing( buttonSpacing );
            break;
          }
      }
  }


  void nVIDIAClient::updateMask()
  {
    if ( !nvidia_initialized )
      return;

    QRegion mask( 0, 0, width(), height() );
    //  QRegion mask(width(), titlebar_->geometry().height());
    int r( width() );
    int b( height()-1 );
    // Remove top-left corner.
    mask -= QRegion( 0, 0, 4, 1 );
    mask -= QRegion( 0, 1, 3, 1 );
    mask -= QRegion( 0, 2, 2, 1 );
    mask -= QRegion( 0, 3, 1, 1 );
    // Remove top-right corner.
    mask -= QRegion( r - 12, 0, 12, 1 );
    mask -= QRegion( r - 9, 1, 9, 1 );
    mask -= QRegion( r - 7, 2, 7, 1 );
    mask -= QRegion( r - 6, 3, 6, 1 );
    mask -= QRegion( r - 5, 4, 5, 1 );
    mask -= QRegion( r - 4, 5, 4, 1 );
    mask -= QRegion( r - 3, 6, 3, 2 );
    mask -= QRegion( r - 2, 8, 2, 3 );
    mask -= QRegion( r - 1, 11, 1, 13 );
    // Remove bottom-left corner.
    mask -= QRegion(0, b-0, 2, b-1);
    mask -= QRegion(0, b-1, 1, b-1);
    // Remove bottom-right corner.
    mask -= QRegion(r - 2, b-0, 2, b-1);
    mask -= QRegion(r - 1, b-1, 1, b-1);
    //apply mask
    setMask( mask );
    maskDirty = false;
  }


  void nVIDIAClient::updateCaptionBuffer()
  {
    if ( !nvidia_initialized )
      return;

    bool active = isActive();
    QPixmap *icon = NULL;

    if ( captionBuffer.size() != captionRect.size() )
      captionBuffer.resize( captionRect.size() );

    if ( captionBuffer.isNull() )
      return;

    QPainter p( &captionBuffer );

    // Draw the caption bubble
    if ( active && largeCaption )
      {
        p.drawPixmap( 0, 0, *clientHandler->tile( CaptionLargeLeft, true ) );
        p.drawTiledPixmap( 15, 0, captionRect.width() - 30, captionRect.height(),
                           *clientHandler->tile( CaptionLargeCenter, true ) );
        p.drawPixmap( captionRect.width() - 15, 0, *clientHandler->tile( CaptionLargeRight, true ) );
      }
    else
      {
        p.drawPixmap( 0, 0, *clientHandler->tile( CaptionSmallLeft, active ) );
        p.drawTiledPixmap( 15, 0, captionRect.width() - 30, captionRect.height(),
                           *clientHandler->tile( CaptionSmallCenter, active ) );
        p.drawPixmap( captionRect.width() - 15, 0, *clientHandler->tile( CaptionSmallRight, active ) );
      }

    if ( clientHandler->showAppIcons() )
      {
        if ( active )
          {
            if ( ! activeIcon )
              activeIcon = new QPixmap( this->icon().pixmap( QIconSet::Small, QIconSet::Normal )); // FRAME
            icon = activeIcon;
          }
        else
          {
            if ( ! inactiveIcon )
              {
                QImage img = this->icon().pixmap( QIconSet::Small, QIconSet::Normal ).convertToImage();
                KIconEffect::semiTransparent( img );
                inactiveIcon = new QPixmap( img );
              }
            icon = inactiveIcon;
          }
      }

    p.setFont( options()->font( active ) );
    int tw = p.fontMetrics().width( caption() ) +
             ( clientHandler->showAppIcons() ? 16 + iconSpacing : 0 );

    int xpos = QMAX( (captionRect.width() - tw) / 3, 8 );
    QRect tr = QStyle::visualRect( QRect(1, 1, captionRect.width() - xpos - 10,
                                         captionRect.height() - 1), captionBuffer.rect() );

    //p.setPen( Qt::red ); // debug
    //p.drawRect( tr );    // debug

    // Application icon
    if ( clientHandler->showAppIcons() )
      {
        QRect iconRect = QStyle::visualRect( QRect(tr.x(), 1 + (captionRect.height() - 16) / 2, 16, 16), tr );
        QRect r( icon->rect() );
        r.moveCenter( iconRect.center() );

        if ( tr.width() > 16 )
          {
            p.drawPixmap( r, *icon );
          }
        else
          {
            QRect sr( 0, 0, icon->width(), icon->height() );

            if ( QApplication::reverseLayout() )
              sr.addCoords( icon->width() - tr.width(), 0, 0, 0 );
            else
              sr.addCoords( 0, 0, -( icon->width() - tr.width() ), 0 );

            p.drawPixmap( r.x() + sr.x(), r.y() + sr.y(), *icon,
                          sr.x(), sr.y(), sr.width(), sr.height() );
          }

        //p.drawRect( r ); // debug

        // if ( QApplication::reverseLayout() )
        //tr.addCoords( 0, 0, -(16 + iconSpacing), 0 );
        // else
        tr.addCoords( 0, 0, 0, 0 );
        tr.addCoords( (16 + iconSpacing), 0, 0, 0 );
      }

    // Draw the titlebar text
    int flags = AlignLeft | AlignVCenter | SingleLine;
    // flags |= ( QApplication::reverseLayout() ? AlignRight : AlignLeft );

    if ( clientHandler->useShadowedText() )
      {
        p.translate( QApplication::reverseLayout() ? -1 : 1, 1 );
        p.setPen(QColor(213, 213, 221)); //options()->color(ColorTitleBar, active).dark() );
        // p.setPen( black );
        p.drawText( tr, flags, caption() );
        p.translate( QApplication::reverseLayout() ? 1 : -1, -1 );
      }

    p.setPen( options()->color( ColorFont, active ) );
    p.drawText( tr, flags, caption() );

    captionBufferDirty = false;
  }


  void nVIDIAClient::calculateCaptionRect()
  {
    QFontMetrics fm( options()->font(isActive()) );
    int cw = fm.width( caption() ) + 95;
    int titleBaseY = ( largeTitlebar ? 3 : 0 );

    if ( clientHandler->showAppIcons() )
      cw += 16 + 4; // icon width + space

    cw = QMIN( cw, titlebar->geometry().width() );
    captionRect = QStyle::visualRect( QRect(titlebar->geometry().x(), (largeCaption ? 0 : titleBaseY),
                                            cw, clientHandler->titleBarHeight(largeCaption) ),
                                      titlebar->geometry() );
  }


  void nVIDIAClient::captionChange()
  {
    QRect r( captionRect );
    calculateCaptionRect();

    if ( r.size() != captionRect.size() )
      maskDirty = true;

    captionBufferDirty = true;

    widget()->repaint( r | captionRect, false );
  }


  void nVIDIAClient::iconChange()
  {
    if ( clientHandler->showAppIcons() )
      {

        // Force updateCaptionBuffer() to recreate the cached icons
        if ( activeIcon )
          delete activeIcon;

        if ( inactiveIcon )
          delete inactiveIcon;

        activeIcon = inactiveIcon = NULL;

        captionBufferDirty = true;
        widget()->repaint( captionRect, false );
      }
  }


  void nVIDIAClient::activeChange()
  {
    bool active = isActive();

    // Note: It's assumed that the same font will always be used for both active
    //       and inactive windows, since the fonts kcm hasn't supported setting
    //       different fonts for different window states for some time.
    if ( largeTitlebar )
      {
        largeCaption = ( active && !maximizedVertical() );
        calculateCaptionRect();
        maskDirty = true;
      }

    captionBufferDirty = true;

    widget()->repaint( false );

    for ( int i=0; i < NumButtons; i++ )
      if ( button[i] )
        button[i]->repaint( false );
  }


  void nVIDIAClient::maximizeChange()
  {
    if ( clientHandler->largeCaptionBubbles() )
      {
        if ( maximizeMode() & MaximizeVertical )
          {
            // We've been maximized - shrink the titlebar by 3 pixels
            topSpacer->changeSize( 10, 1, QSizePolicy::Expanding, QSizePolicy::Minimum );
            largeCaption = largeTitlebar = false;

            calculateCaptionRect();
            captionBufferDirty = maskDirty = true;

            widget()->layout()->activate();
            widget()->repaint( false );
          }
        else if (( maximizeMode() & MaximizeVertical ) == 0 && !largeTitlebar )
          {
            // We've been restored - enlarge the titlebar by 3 pixels
            topSpacer->changeSize( 10, 4, QSizePolicy::Expanding, QSizePolicy::Minimum );
            largeCaption = largeTitlebar = true;

            calculateCaptionRect();
            captionBufferDirty = maskDirty = true;

            widget()->layout()->activate();
            widget()->repaint( false );
          }
      }

    if ( button[ MaxButton ] )
      {
        QToolTip::remove
          ( button[ MaxButton ] );
        QToolTip::add
          ( button[ MaxButton ], maximizeMode() == MaximizeFull ? i18n("Restore") : i18n("Maximize") );
        button[ MaxButton ]->repaint();
      }
  }


  void nVIDIAClient::desktopChange()
  {
    if ( button[ OnAllDesktopsButton ] )
      {
        QToolTip::remove
          ( button[ OnAllDesktopsButton ] );
        QToolTip::add
          ( button[ OnAllDesktopsButton ], isOnAllDesktops() ? i18n("Not On All Desktops") : i18n("On All Desktops") );
      }
  }


  void nVIDIAClient::menuButtonPressed()
  {
    QPoint menuPoint ( button[MenuButton]->rect().bottomLeft().x() - 6,
                       button[MenuButton]->rect().bottomLeft().y() + 3 );
    KDecorationFactory* f = factory();
    showWindowMenu( button[MenuButton]->mapToGlobal( menuPoint ));
    if( !f->exists( this )) // 'this' was destroyed
      return;
    button[MenuButton]->setDown(false);
  }


  void nVIDIAClient::slotMaximize()
  {
    switch ( button[ MaxButton ]->lastButton() )
      {
      case MidButton:
        maximize( maximizeMode() ^ MaximizeVertical );
        break;

      case RightButton:
        maximize( maximizeMode() ^ MaximizeHorizontal );
        break;

      case LeftButton:
        maximize( maximizeMode() == MaximizeFull ? MaximizeRestore : MaximizeFull );
        break;
      }
  }


  void nVIDIAClient::paintEvent( QPaintEvent *e )
  {
    if ( !nvidia_initialized )
      return;

    QPainter p( widget());
    QRect updateRect( e->rect() );
    bool active = isActive();

    int titleBaseY         = ( largeTitlebar ? 3 : 0 );
    int titleBarHeight     = clientHandler->titleBarHeight( largeTitlebar );
    int grabBarHeight      = clientHandler->grabBarHeight();
    int leftBorderWidth    = clientHandler->tile( BorderLeft, active )->width();
    int rightBorderWidth   = clientHandler->tile( BorderRight, active )->width();

    if ( maskDirty )
      updateMask();

    // Titlebar
    // -----------------------------------------------------------------------
    if ( updateRect.y() < titleBarHeight )
      {
        int titleBarBaseHeight = titleBarHeight - titleBaseY;

        if ( captionBufferDirty )
          updateCaptionBuffer();

        // Top left corner
        if ( updateRect.x() < clientHandler->tile( TitleLeft, active )->width() )
          p.drawPixmap( 0, titleBaseY,
                        *clientHandler->tile( TitleLeft, active ) );

        // Space between the top left corner and the caption bubble
        if ( updateRect.x() < captionRect.left() && updateRect.right() >= clientHandler->tile( TitleLeft, active )->width() )
          {
            int x1 = QMAX( clientHandler->tile( TitleLeft, active )->width(), updateRect.x() );
            int x2 = QMIN( captionRect.left(), updateRect.right() );

            p.drawTiledPixmap( x1, titleBaseY, x2 - x1 + 1, titleBarBaseHeight,
                               *clientHandler->tile( TitleCenter, active ) );
          }

        // Caption bubble
        if ( updateRect.x() <= captionRect.right() && updateRect.right() > 15 )
          {
            if ( captionRect.width() >= 25 )
              p.drawPixmap(clientHandler->tile( TitleLeft, active )->width(), active ? 0 : titleBaseY, captionBuffer );
            else
              p.drawTiledPixmap( captionRect.x(), titleBaseY, captionRect.width(),
                                 titleBarBaseHeight, *clientHandler->tile( TitleCenter, active ) );
          }

        // Space between the caption bubble and the top right corner
        if ( updateRect.right() > captionRect.right() && updateRect.x() < width() - clientHandler->tile( TitleRight, active )->width() )
          { // FRAME
            int x1 = QMAX( captionRect.right() + 1, updateRect.x() );
            int x2 = QMIN( width() - clientHandler->tile( TitleRight, active )->width(), updateRect.right() );

            p.drawTiledPixmap( x1, titleBaseY, x2 - x1 + 1, titleBarBaseHeight,
                               *clientHandler->tile( TitleCenter, active ) );
          }

        // Top right corner
        if ( updateRect.right() >= width() - clientHandler->tile( TitleRight, active )->width() )
          p.drawPixmap( width() - clientHandler->tile( TitleRight, active )->width(), titleBaseY,
                        *clientHandler->tile( TitleRight, active ) );
      }

    // Borders
    // -----------------------------------------------------------------------
    if ( updateRect.bottom() >= titleBarHeight &&
         updateRect.top() < height() - grabBarHeight )
      {
        int top    = QMAX( titleBarHeight, updateRect.top() );
        int bottom = QMIN( updateRect.bottom(), height() - grabBarHeight );

        // Left border
        if ( updateRect.x() < leftBorderWidth )
          p.drawTiledPixmap( 0, top, leftBorderWidth, bottom - top + 1,
                             *clientHandler->tile( BorderLeft, active ) );

        // Right border
        if ( e->rect().right() > width() - rightBorderWidth - 1 )
          p.drawTiledPixmap( width() - rightBorderWidth, top, rightBorderWidth,
                             bottom - top + 1, *clientHandler->tile( BorderRight, active ) );
      }

    // Bottom grab bar
    // -----------------------------------------------------------------------
    if ( updateRect.bottom() >= height() - grabBarHeight )
      {
        // Bottom left corner
        if ( updateRect.x() < 9 )
          p.drawPixmap( 0, height() - grabBarHeight,
                        *clientHandler->tile( GrabBarLeft, active ) );

        // Space between the left corner and the right corner
        if ( updateRect.x() < width() - 9 )
          {
            int x1 = QMAX( 9, updateRect.x() );
            int x2 = QMIN( width() - 9, updateRect.right() );

            p.drawTiledPixmap( x1, height() - grabBarHeight, x2 - x1 + 1,
                               grabBarHeight, *clientHandler->tile( GrabBarCenter, active ) );
          }

        // Bottom right corner
        if ( updateRect.right() > width() - 9 )
          p.drawPixmap( width() - 9, height() - grabBarHeight,
                        *clientHandler->tile( GrabBarRight, active ) );
      }

    // Extra drawline for the 1 pixel empty space QLayout leaves when a window is shaded.
    /*p.setPen( options()->color( ColorTitleBlend, active ) );
    p.drawLine( leftBorderWidth, height() - grabBarHeight - 1,
    			width() - rightBorderWidth - 1, height() - grabBarHeight - 1 );*/
  }


  void nVIDIAClient::resizeEvent( QResizeEvent *e )
  {
    // FRAME	Client::resizeEvent( e );

    QRect r( captionRect );
    calculateCaptionRect();

    if ( r.size() != captionRect.size() )
      captionBufferDirty = true;

    maskDirty = true;

    if ( widget()->isVisible() )
      {
        widget()->update( widget()->rect() );
        int dx = 0;
        int dy = 0;

        if ( e->oldSize().width() != width() )
          dx = 32 + QABS( e->oldSize().width() -  width() );

        if ( e->oldSize().height() != height() )
          dy = 8 + QABS( e->oldSize().height() -  height() );

        if ( dy )
          widget()->update( 0, height() - dy + 1, width(), dy );

        if ( dx )
          {
            widget()->update( width() - dx + 1, 0, dx, height() );
            widget()->update( QRect( QPoint(4,4), titlebar->geometry().bottomLeft() - QPoint(1,0) ) );
            widget()->update( QRect( titlebar->geometry().topRight(), QPoint( width() - 4,
                                     titlebar->geometry().bottom() ) ) );
            // Titlebar needs no paint event
            QApplication::postEvent( this, new QPaintEvent( titlebar->geometry(), FALSE ) );
          }
      }
  }


  void nVIDIAClient::mouseDoubleClickEvent( QMouseEvent *e )
  {
    if ( QRect( 0, 0, width(), clientHandler->titleBarHeight( largeTitlebar ) ).contains( e->pos() ) )
      titlebarDblClickOperation();
  }


  nVIDIAClient::Position nVIDIAClient::mousePosition( const QPoint &p ) const
    {
      int titleBaseY = (largeTitlebar ? 3 : 0);

      int leftBorder   = clientHandler->tile( BorderLeft, true )->width();
      int rightBorder  = width() - clientHandler->tile( BorderRight, true )->width() - 1;
      int bottomBorder = height() - clientHandler->grabBarHeight() - 1;
      int bottomCornerSize = 3*clientHandler->tile( BorderRight, true )->width()/2 + 24;

      // Test if the mouse is over the titlebar area
      if ( p.y() < titleBaseY + 11 )
        {
          // Test for the top left corner
          if ( p.x() < leftBorder + 11 )
            {
              if ( (p.y() < titleBaseY + 3 && p.x() < leftBorder + 11) ||
                   (p.y() < titleBaseY + 6 && p.x() < leftBorder + 6) ||
                   (p.y() < titleBaseY + 11 && p.x() < leftBorder + 3) )
                return PositionTopLeft;
            }

          // Test for the top right corner
          if ( p.x() > rightBorder - 11 )
            {
              if ( (p.y() < titleBaseY + 3 && p.x() > rightBorder - 11) ||
                   (p.y() < titleBaseY + 6 && p.x() > rightBorder - 6) ||
                   (p.y() < titleBaseY + 11 && p.x() > rightBorder - 3) )
                return PositionTopRight;
            }

          // Test for the top border
          if ( p.y() <= 3 || (p.y() <= titleBaseY+3 &&
                              (p.x() < captionRect.left() || p.x() > captionRect.right()) ) )
            return PositionTop;

          // The cursor must be over the center of the titlebar.
          return PositionCenter;
        }

      // Test the sides
      else if ( p.y() < bottomBorder )
        {
          // Test for the left side
          if ( p.x() < leftBorder )
            {
              if ( p.y() < height() - bottomCornerSize )
                return PositionLeft;
              else
                return PositionBottomLeft;
            }

          // Test for the right side
          else if ( p.x() > rightBorder )
            {
              if ( p.y() < height() - bottomCornerSize )
                return PositionRight;
              else
                return PositionBottomRight;
            }

          // The cursor must be over the center of the window
          return PositionCenter;
        }

      // Test the grab bar / bottom border
      else
        {
          // Test for the bottom left corner
          if ( p.x() < bottomCornerSize )
            return PositionBottomLeft;

          // Test for the bottom right corner
          else if ( p.x() > width() - bottomCornerSize - 1 )
            return PositionBottomRight;

          // The cursor must be over the bottom border
          return PositionBottom;
        }

      // We should never get here
      return PositionCenter;
    }


  void nVIDIAClient::resize( const QSize& s )
  {
    widget()->resize( s );
  }


  void nVIDIAClient::borders( int& left, int& right, int& top, int& bottom ) const
    {
      int titleBarHeight     = clientHandler->titleBarHeight( clientHandler->largeCaptionBubbles() );
      int grabBarHeight      = clientHandler->grabBarHeight();
      int leftBorderWidth    = clientHandler->tile( BorderLeft, isActive() )->width();
      int rightBorderWidth   = clientHandler->tile( BorderRight, isActive() )->width();

      left   = leftBorderWidth;
      right  = rightBorderWidth;
      top    = titleBarHeight;
      bottom = grabBarHeight;

      if ( ( maximizeMode() & MaximizeHorizontal ) && !options()->moveResizeMaximizedWindows())
        left = right = 0;
      if( maximizeMode() & MaximizeVertical)
        {
          top = clientHandler->titleBarHeight( false );
          if( !options()->moveResizeMaximizedWindows())
            bottom = 0;
        }
    }


  QSize nVIDIAClient::minimumSize() const
    {
      return widget()->minimumSize();
    }


  bool nVIDIAClient::eventFilter( QObject* o, QEvent* e )
  {
    if ( o != widget() )
      return false;

    switch ( e->type() )
      {
      case QEvent::Resize:
        resizeEvent( static_cast< QResizeEvent* >( e ) );
        return true;

      case QEvent::Paint:
        paintEvent( static_cast< QPaintEvent* >( e ) );
        return true;

      case QEvent::MouseButtonDblClick:
        mouseDoubleClickEvent( static_cast< QMouseEvent* >( e ) );
        return true;

      case QEvent::MouseButtonPress:
        processMousePressEvent( static_cast< QMouseEvent* >( e ) );
        return true;

      default:
        return false;
      }
  }

} // namespace nVIDIA



// -------------------------------------------------------------------------------------------



extern "C"
  {
    KDecorationFactory *create_factory()
    {
      nVIDIA::clientHandler = new nVIDIA::nVIDIAHandler();
      return nVIDIA::clientHandler;
    }
  }



// vim: set noet ts=4 sw=4:
